from math import inf
from typing import Dict, Hashable, List, NamedTuple, Tuple
from networkx import DiGraph

from generalized_geography.common.constants import LOOP, LOSE, UNKNOWN, WIN
from generalized_geography.types.bipartite import BipartiteNode
from generalized_geography.types.node_types import NodeType


def get_single_out_edges(graph: DiGraph) -> List[Tuple[Hashable, Hashable]]:
    single_out_edge_nodes = [
        u for u, deg in graph.out_degree() if u.pos == 0 and deg == 1]

    return [(u, v) for u in single_out_edge_nodes for v in graph.successors(u)]


def get_single_in_edges(graph: DiGraph) -> List[Tuple[Hashable, Hashable]]:
    single_in_edge_nodes = [
        v for v, deg in graph.in_degree() if v.pos == 1 and deg == 1]
    return [(u, v) for v in single_in_edge_nodes for u in graph.predecessors(v)]


def get_all_inferior_edges(graph: DiGraph) -> List[Tuple[Hashable, Hashable]]:
    single_out_edges = get_single_out_edges(graph)
    single_in_edges = get_single_in_edges(graph)

    return list(set(single_out_edges) | set(single_in_edges))


def remove_two_cycles(graph: DiGraph, weight_key: str = "num") -> List[Tuple[BipartiteNode, BipartiteNode, int]]:

    inferior_edges = get_all_inferior_edges(graph)
    to_remove = []

    for u, v in inferior_edges:
        if graph.has_edge(v, u):
            m = graph[v][u][weight_key]
            if m < 2:
                continue
            to_remove.append((v, u, m - m % 2))
    counter = {}
    for i in range(len(inferior_edges) - 1):
        start1, end1 = inferior_edges[i]
        for j in range(i + 1, len(inferior_edges)):
            start2, end2 = inferior_edges[j]

            if graph.has_edge(end1, start2) and graph.has_edge(end2, start1):
                num1 = graph[end1][start2][weight_key] - \
                    counter.get((end1, start2), 0)
                num2 = graph[end2][start1][weight_key] - \
                    counter.get((end2, start1), 0)
                if num1 <= 0 or num2 <= 0:
                    continue

                delete_num = min(num1, num2)
                counter[(end1, start2)] = counter.get(
                    (end1, start2), 0) + delete_num
                counter[(end2, start1)] = counter.get(
                    (end2, start1), 0) + delete_num
                to_remove.append((end1, start2, delete_num))
                to_remove.append((end2, start1, delete_num))

    for u, v, m in to_remove:
        if graph[u][v][weight_key] <= m:
            graph.remove_edge(u, v)
        else:
            graph[u][v][weight_key] -= m

    return to_remove


class CdegClassifyFastResult(NamedTuple):
    pruned_graph: DiGraph
    type_map: Dict[BipartiteNode, int]
    two_cycles: List[Tuple[BipartiteNode, BipartiteNode, int]]


def cdeg_classify_fast_once(graph: DiGraph, inplace: bool = False, weight_key: str = "num") -> CdegClassifyFastResult:
    if not inplace:
        graph = graph.copy()
    loop_map = {}
    stack = []
    type_map: Dict[BipartiteNode, NodeType] = {}
    two_cycles = remove_two_cycles(graph, weight_key=weight_key)

    def get_seed_type(node: BipartiteNode) -> NodeType:
        if node.pos == 0:
            if graph.out_degree(node) == 0:
                return LOSE
            else:
                return UNKNOWN
        else:
            outdeg = graph.out_degree(node)
            if outdeg >= 2:
                return UNKNOWN
            elif outdeg == 0:
                return LOSE
            else:
                succ = list(graph.successors(node))[0]
                if graph.out_degree(succ) != 1 or list(graph.successors(succ))[0] != node:
                    return UNKNOWN

                num = graph[node][succ][weight_key]
                if num % 2 == 1:
                    loop_map[node] = succ
                    return LOOP

    def set_seeds():
        for node in list(graph.nodes):
            nodeType = get_seed_type(node)
            if nodeType != UNKNOWN:
                type_map[node] = nodeType
                stack.append(node)

    set_seeds()

    while stack:
        node = stack.pop()
        preds = [pred for pred in graph.predecessors(
            node) if pred not in type_map]

        type = type_map[node]
        graph.remove_node(node)
        if node.pos == 0:
            if type == LOSE:
                for pred in preds:
                    type_map[pred] = WIN
                    stack.append(pred)
            else:
                for pred in preds:
                    pred_type = get_seed_type(pred)
                    if pred_type != UNKNOWN:
                        type_map[pred] = pred_type
                        stack.append(pred)
        else:
            if type == LOSE:
                for pred in preds:
                    pred_type = get_seed_type(pred)
                    if pred_type != UNKNOWN:
                        type_map[pred] = LOSE
                        stack.append(pred)
            else:
                for pred in preds:
                    type_map[pred] = type
                    stack.append(pred)
    return CdegClassifyFastResult(graph, type_map, two_cycles)


def cdeg_classify_fast(graph: DiGraph, inplace: bool = False, weight_key: str = "num") -> CdegClassifyFastResult:
    if not inplace:
        graph = graph.copy()

    total_two_cycles = []
    while True:
        result = cdeg_classify_fast_once(
            graph, inplace=True, weight_key=weight_key)
        total_two_cycles.extend(result.two_cycles)
        if len(result.pruned_graph.nodes) == len(graph.nodes):
            break
        graph = result.pruned_graph

    return CdegClassifyFastResult(graph, result.type_map, total_two_cycles)


if __name__ == "__main__":

    with open("dataset/oldict.txt", "r", encoding="UTF-8") as f:
        words = [word.strip() for word in f.read().split("\n")]



# Directed Vertex Geography


from typing import Dict, Hashable, List, NamedTuple, Tuple
from networkx import DiGraph, descendants

from generalized_geography.common.constants import WIN
from generalized_geography.solver.rdg import rdg_classify


def remove_loops(graph: DiGraph) -> List[Tuple[Hashable, Hashable]]:
    loop_edges = [(node, node) for node in graph if graph.has_edge(node, node)]
    for node in graph:
        if graph.has_edge(node, node):
            graph.remove_edge(node, node)
    return loop_edges


class ClassifyDvgResult(NamedTuple):
    pruned_graph: DiGraph
    type_map: Dict[Hashable, int]
    loop_edges: List[Tuple[Hashable, Hashable]]


def dvg_classify_fast(graph: DiGraph, inplace: bool = False) -> ClassifyDvgResult:
    if not inplace:
        graph = graph.copy()
    loop_edges = remove_loops(graph)
    result = rdg_classify(graph, inplace=True)

    return ClassifyDvgResult(pruned_graph=graph, type_map=result.type_map, loop_edges=loop_edges)


def dvg_is_win(graph: DiGraph, start_node: Hashable) -> bool:

    graph = graph.copy()
    result = dvg_classify_fast(graph, inplace=True)

    if start_node in result.type_map:
        return result.type_map[start_node] == WIN

    reachable_nodes = descendants(graph, start_node)
    subgraph_view = graph.subgraph(reachable_nodes) | {start_node}
    for succ in graph.successors(start_node):
        subgraph: DiGraph = subgraph_view.copy()
        subgraph.remove_node(start_node)
        if not dvg_is_win(subgraph, succ):
            return True
    return False


def dvg_classify_complete(graph: DiGraph, inplace: bool = False) -> Dict[Hashable, bool]:
    type_map = {}
    if not inplace:
        graph = graph.copy()
    result = dvg_classify_fast(graph, inplace=True)

    for node in graph.nodes:
        if node in result.type_map:
            type_map[node] = result.type_map[node] == WIN
        else:
            type_map[node] = dvg_is_win(graph, node)
    return type_map


if __name__ == "__main__":
    g = DiGraph()
    g.add_edge(1, 2)
    g.add_edge(2, 3)
    g.add_edge(3, 1)
    g.add_edge(2, 4)
    g.add_edge(4, 5)
    g.add_edge(5, 6)
    g.add_edge(6, 4)

    print(dvg_classify_fast(g))
    print(dvg_classify_complete(g))



from importlib import resources
from typing import List


def load_oldict_words() -> List[str]:
    """
    패키지 내 datasets/oldict.txt를 읽어서 각 줄을 리스트로 반환
    """
    with resources.open_text("generalized_geography.datasets", "oldict.txt", encoding="utf-8") as f:
        return [line.strip() for line in f if line.strip()]



__version__ = '0.0.11'

from . import solver
from .solver import *

from . import common
from .common import *

from . import types
from .types import *

from . import datasets
from .datasets import *



from typing import List
import hgtk
from networkx import DiGraph
from generalized_geography.common.utils import nodes_in_pos, split_nodes_by_type
from generalized_geography.datasets.oldict import load_oldict_words
from generalized_geography.solver.changeable_deg import cdeg_classify_fast
from generalized_geography.types.bipartite import BipartiteNode


def std(char: str):
    cho, jung, jong = hgtk.letter.decompose(char)
    if cho == "ㄹ" and jung in ["ㅑ", "ㅕ", "ㅛ", "ㅠ", "ㅣ", "ㅖ"]:
        return [char, hgtk.letter.compose("ㅇ", jung, jong)]

    elif cho == "ㄹ" and jung in ["ㅏ", "ㅐ", "ㅗ", "ㅜ", "ㅡ", "ㅚ"]:
        return [char, hgtk.letter.compose("ㄴ", jung, jong)]

    elif cho == "ㄴ" and jung in ["ㅕ", "ㅛ", "ㅠ", "ㅣ"]:
        return [char, hgtk.letter.compose("ㅇ", jung, jong)]

    else:
        return [char]


def words_to_graph(words: List[str]) -> DiGraph:

    g = DiGraph()
    for word in words:
        start_node = BipartiteNode(word[0], 1)
        end_node = BipartiteNode(word[-1], 0)
        if g.has_edge(start_node, end_node):
            g[start_node][end_node]["num"] += 1
        else:
            g.add_edge(start_node, end_node, num=1)

    for node in g.nodes:
        if node.pos == 0:

            for changeable in std(node.name):

                if BipartiteNode(changeable, 1) not in g.nodes:
                    continue

                next_node = BipartiteNode(changeable, 1)
                g.add_edge(node, next_node)

    return g

def load_guel_graph():
    words = load_oldict_words()
    graph = words_to_graph(words)
    return graph





# Directed Edge/Vertex Geography with Repetitions

from typing import Dict, Hashable, Tuple, NamedTuple
from networkx import DiGraph

from generalized_geography.common.constants import LOSE, WIN


class ClassifyRdgResult(NamedTuple):
    pruned_graph: DiGraph
    type_map: Dict[Hashable, int]


def rdg_classify(graph: DiGraph, inplace: bool = False) -> ClassifyRdgResult:
    if not inplace:
        graph = graph.copy()
    sinks = [node for node, deg in graph.out_degree() if deg == 0]
    type_map = {node: LOSE for node in sinks}
    while sinks:
        node = sinks.pop()
        preds = [pred for pred in graph.predecessors(
            node) if pred not in type_map]
        graph.remove_node(node)
        if type_map[node] == LOSE:
            for pred in preds:
                type_map[pred] = WIN
                sinks.append(pred)
        else:
            for pred in preds:
                if graph.out_degree(pred) == 0:
                    type_map[pred] = LOSE
                    sinks.append(pred)
    return ClassifyRdgResult(pruned_graph=graph, type_map=type_map)


if __name__ == "__main__":
    import networkx as nx
    G = nx.DiGraph()
    G.add_edges_from([(1, 2), (2, 3), (3, 1), (3, 4)])
    result = rdg_classify(G)



from typing import List, Literal
from generalized_geography.common.constants import LOOP, LOSE, WIN
from generalized_geography.types.bipartite import BipartiteNode
from generalized_geography.types.node_types import NodeType, ResolvedNodeType


def resolve_node_type(node_type: NodeType) -> ResolvedNodeType:

    if node_type == LOOP:
        return WIN
    else:
        return node_type


def split_nodes_by_type(type_map: dict) -> tuple[List[BipartiteNode], List[BipartiteNode], List[BipartiteNode]]:
    lose_nodes = [node for node, ntype in type_map.items() if ntype == LOSE]
    win_nodes = [node for node, ntype in type_map.items() if ntype == WIN]
    loop_nodes = [node for node, ntype in type_map.items() if ntype == LOOP]

    return lose_nodes, win_nodes, loop_nodes


def split_nodes_by_resolved_type(type_map: dict) -> tuple[List[BipartiteNode], List[BipartiteNode]]:
    lose_nodes = [node for node, ntype in type_map.items(
    ) if resolve_node_type(ntype) == LOSE]
    win_nodes = [node for node, ntype in type_map.items()
                 if resolve_node_type(ntype) == WIN]

    return lose_nodes, win_nodes


def nodes_in_pos(nodes: List[BipartiteNode], pos: Literal[0, 1]) -> List[BipartiteNode]:
    return [node.name for node in nodes if node.pos == pos]



import setuptools

setuptools.setup(
    name="generalized-geography",
    version="0.0.11",
    author="Hyomin",
    author_email="miamiq0000@gmail.com",
    description="Generalized Geography Game Solver",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/singrum/generalized-geography",
    install_requires=['networkx', 'matplotlib'],
    packages=setuptools.find_packages(),
    keywords=['generalized geography', 'word chain', 'game', 'graph theory'],
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    setup_requires=['wheel'],
    python_requires='>=3.6',
)



# Directed Edge Geography


from typing import Dict, Hashable, List, NamedTuple, Tuple
from networkx import DiGraph, descendants
from generalized_geography.common.constants import LOOP, LOSE, WIN
from generalized_geography.common.utils import resolve_node_type
from generalized_geography.types.node_types import NodeType


def remove_2_cycles(graph: DiGraph, weight_key: str = "num") -> List[Tuple[Hashable, Hashable]]:
    to_remove = []

    # self loop 짝수 개 검출
    for node in graph:
        if graph.has_edge(node, node):
            m = graph[node][node][weight_key]
            to_remove.append((node, node, m - m % 2))

    # (a,b), (b,a) 꼴 엣지 검출
    for u, v in [(u, v) for u, v in graph.edges if u > v]:
        if graph.has_edge(v, u):
            delete_num = min(graph[u][v][weight_key],
                             graph[v][u][weight_key])
            to_remove.extend([(u, v, delete_num), (v, u, delete_num)])

    for u, v, m in to_remove:
        graph[u][v][weight_key] -= m
        if graph[u][v][weight_key] <= 0:
            graph.remove_edge(u, v)

    return to_remove


class DegClassifyResult(NamedTuple):
    pruned_graph: DiGraph
    type_map: Dict[Hashable, int]
    two_cycle_edges: List[Tuple[Hashable, Hashable]]


def deg_classify_fast(graph: DiGraph, inplace: bool = False, weight_key: str = "num") -> DegClassifyResult:
    if not inplace:
        graph = graph.copy()
    two_cycle_edges = remove_2_cycles(graph, weight_key=weight_key)

    type_map = {}
    sinks = []
    for node in graph.nodes:
        if graph.out_degree(node) == 0:
            type_map[node] = LOSE
            sinks.append(node)
        elif graph.out_degree(node) == 1 and graph.has_edge(node, node):
            type_map[node] = LOOP
            sinks.append(node)

    while sinks:
        node = sinks.pop()
        preds = list(graph.predecessors(node))
        graph.remove_node(node)
        for pred in preds:
            if pred in type_map:
                continue
            if type_map.get(node) == LOSE:
                type_map[pred] = WIN
                sinks.append(pred)
            else:
                if graph.out_degree(pred) == 0:
                    type_map[pred] = LOSE
                    sinks.append(pred)
                elif graph.out_degree(pred) == 1 and graph.has_edge(pred, pred):
                    type_map[pred] = LOOP
                    sinks.append(pred)

    return DegClassifyResult(pruned_graph=graph, type_map=type_map, two_cycle_edges=two_cycle_edges)


def deg_is_win(graph: DiGraph, start_node: Hashable, weight_key: str = "num") -> bool:
    result = deg_classify_fast(graph, inplace=False, weight_key=weight_key)
    if start_node in result.type_map:
        return resolve_node_type(result.type_map[start_node]) == WIN

    reachable_nodes = descendants(graph, start_node) | {start_node}
    subgraph_view = graph.subgraph(reachable_nodes)

    for succ in list(subgraph_view.successors(start_node)):
        subgraph = subgraph_view.copy()
        subgraph[start_node][succ][weight_key] -= 1
        if subgraph[start_node][succ][weight_key] <= 0:
            subgraph.remove_edge(start_node, succ)

        if not deg_is_win(subgraph, succ, weight_key=weight_key):
            return True

    return False


def deg_classify_complete(graph: DiGraph, inplace: bool = False, weight_key: str = "num") -> Dict[Hashable, NodeType]:
    if not inplace:
        graph_copy = graph.copy()

    result = deg_classify_fast(graph_copy, inplace=True, weight_key=weight_key)

    type_map = {}

    for node in graph.nodes:
        if node in type_map:
            type_map[node] = resolve_node_type(result.type_map[node])
        else:
            type_map[node] = WIN if deg_is_win(
                graph, node, weight_key=weight_key) else LOSE

    return type_map


if __name__ == "__main__":
    pass
